#!/usr/bin/python3
import os
import sys
from PyQt5 import QtCore, QtGui, QtWidgets

HISTORY_FILE = os.path.expanduser("~/.config/obsuite/launcher.hst")


class FramelessWidget(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.setAttribute(QtCore.Qt.WA_X11NetWmWindowTypeSplash)
        self.setWindowFlags(QtCore.Qt.FramelessWindowHint | QtCore.Qt.WindowStaysOnTopHint)

    def mouseMoveEvent(self, event):
        # Mouse dragging for frameless windows
        if event.buttons() == QtCore.Qt.LeftButton:
            self.move(event.globalPos() - self.dragPosition)
            event.accept()

    def mousePressEvent(self, event):
        # Mouse dragging for frameless windows
        if event.button() == QtCore.Qt.LeftButton:
            self.dragPosition = event.globalPos() - self.frameGeometry().topLeft()
            event.accept()

    def paintEvent(self, event):
        # Draw a border for FramelessWindowHint
        borderColor = QtGui.QColor("black")
        bgColor = QtGui.QColor(self.palette().color(QtGui.QPalette.Background))
        painter = QtGui.QPainter(self)
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtGui.QBrush(borderColor))
        painter.drawRect(0, 0, self.width(), self.height())
        painter.setBrush(QtGui.QBrush(bgColor))
        painter.drawRect(1, 1, self.width()-2, self.height()-2)


class Main(FramelessWidget):
    def __init__(self, parent):
        super().__init__()
        self.preferences = parent.preferences
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.slave = QtCore.QProcess()
        self._historyLoad()

        self.cmdLine = CommandLine(self)
        self.cmdLine.textEdited.connect(self.suggestPattern)
        self.cmdLine.execute.connect(self.execute)
        self.cmdLine.close.connect(self.close)
        self.cmdLine.focusOut.connect(self.focusOut)
        self.cmdLine.focusSwitch.connect(self.focusSwitch)
        self.cmdLine.clearRecent.connect(self.clearRecent)
        self.recentList = RecentList(self)
        self.recentList.execute.connect(self.execute)
        self.recentList.focusOut.connect(self.focusOut)
        self.recentList.removeSuggestion.connect(self.removeSuggestion)
        self.recentList.resize.connect(self.resizeToContent)
        self.recentList.focusCmd.connect(self.cmdLine.setFocus)
        self.recentList.deselect.connect(self.cmdLine.deselect)
        self.recentList.currentRowChanged.connect(self.cmdLine.updateText)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(15, 18, 15, 18)
        self.layout.addWidget(self.cmdLine)
        self.layout.addWidget(self.recentList)
        self.setLayout(self.layout)

        self.setMinimumWidth(410)
        self.resize(0, 0)

        if self.preferences.get("parameters", "launcher", "startup suggestions"):
            self._suggestRecent()
        self.show()
        self.activateWindow()

    def _aliasFetch(self):
        alias = {}
        with open(os.path.expanduser("~/.bashrc")) as f:
            for line in f.readlines():
                if line.startswith("alias "):
                    line = line[len("alias "):].strip()
                    line = line.split("=", maxsplit=1)
                    cmd = os.path.expanduser(line[1][1:-1])
                    alias[line[0]] = cmd
        return alias

    def _autoComplete(self, pattern):
        for cmd in self.history:
            if cmd.startswith(pattern):
                length = len(cmd) - len(pattern)
                self.cmdLine.setText(cmd)
                self.cmdLine.setSelection(len(pattern), length)
                break

    def _deleteFromHistory(self, cmd):
        for line in list(self.history):
            if line == cmd:
                self.history.remove(line)

    def _historyLoad(self):
        self.history = []
        if not os.path.isfile(HISTORY_FILE):
            with open(HISTORY_FILE, "a") as f:
                f.close()
        with open(HISTORY_FILE) as f:
            for line in f.readlines():
                self.history.append(line.strip())
        self.history = list(reversed(self.history))

    def _parse(self, cmd):
        cmd = cmd.replace("~", os.path.expanduser("~"))
        cmd = cmd.split(" ")
        fetch = self._aliasFetch()
        for alias in fetch:
            if alias == cmd[0]:
                cmd[0] = fetch[alias]
        return " ".join(cmd)

    def _save(self, cmd):
        self._deleteFromHistory(cmd)
        self.history.insert(0, cmd)
        with open(HISTORY_FILE, "w") as f:
            for line in reversed(self.history):
                f.write(f"{line}\n")

    def _search(self, pattern):
        items = []
        maximum = self.preferences.get("parameters", "launcher", "suggestions amount")
        for index, cmd in enumerate(self.history):
            if len(items) == maximum: break
            if index == 0 and not self.cmdLine.backspace:
                self._autoComplete(pattern)
            elif cmd.startswith(pattern):
                if not cmd == self.cmdLine.text():
                    items.append(cmd)
        self.recentList.addItems(items)

    def _setErrorPalette(self, sender):
        fg = QtGui.QColor("#ffc7ce")
        bg = QtGui.QColor("#9c0006")
        palette = QtGui.QPalette()
        if isinstance(sender, QtWidgets.QListWidgetItem):
            sender.setForeground(fg)
            sender.setBackground(bg)
            palette.setColor(QtGui.QPalette.Highlight, bg)
            palette.setColor(QtGui.QPalette.HighlightedText, fg)
            self.recentList.setPalette(palette)
        else:
            palette.setColor(QtGui.QPalette.Base, bg)
            palette.setColor(QtGui.QPalette.Text, fg)
            sender.setPalette(palette)

    def _suggestRecent(self):
        self.recentList.clear()
        maximum = self.preferences.get("parameters", "launcher", "suggestions amount")
        for lastest in self.history:
            if self.recentList.count() == maximum: break
            if not lastest == self.cmdLine.text():
                self.recentList.addItem(lastest)
        self.resizeToContent()

    def clearRecent(self):
        self.recentList.hide()
        self.resizeToContent()

    def execute(self, sender):
        raw = sender.text().strip()
        cmd = self._parse(raw)
        if not cmd:
            self.close()
        elif self.slave.startDetached(cmd):
            self._save(raw)
            self.close()
        else:
            self._setErrorPalette(sender)

    def focusOut(self):
        if self.preferences.get("parameters", "launcher", "close on focus out"):
            if not self.cmdLine.hasFocus() and not self.recentList.hasFocus():
                self.close()

    def focusSwitch(self, row):
        self.recentList.setFocus(True)
        self.recentList.setCurrentRow(row)

    def removeSuggestion(self, cmd):
        self._deleteFromHistory(cmd)
        with open(HISTORY_FILE, "w") as f:
            for line in reversed(self.history):
                f.write(f"{line}\n")

    def resizeToContent(self):
        self.recentList.resizeToContent()
        QtCore.QCoreApplication.processEvents()
        self.resize(0, 0)

    def suggestPattern(self, pattern):
        if pattern:
            self.recentList.clear()
            self._search(pattern)
            self.resizeToContent()
        else:
            self._suggestRecent()


class CommandLine(QtWidgets.QLineEdit):
    execute = QtCore.pyqtSignal(object)
    focusSwitch = QtCore.pyqtSignal(int)
    focusOut = QtCore.pyqtSignal()
    clearRecent = QtCore.pyqtSignal()
    close = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.index = 0
        self.setTextMargins(5, 2, 5, 2)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFocus(True)
        if parent.history:
            self.setText(parent.history[0])
            self.selectAll()

    def focusOutEvent(self, event):
        self.focusOut.emit()

    def keyReleaseEvent(self, event):
        if event.key() == QtCore.Qt.Key_Backspace:
            self.backspace = False
        QtWidgets.QLineEdit.keyReleaseEvent(self, event)

    def keyPressEvent(self, event):
        self._setDefaultPalette()
        key = event.key()

        if key == QtCore.Qt.Key_Escape:
            self.close.emit()

        elif key == QtCore.Qt.Key_Return or key == QtCore.Qt.Key_Enter:
            self.execute.emit(self)

        elif key == QtCore.Qt.Key_Up:
            if self.parent.recentList.isVisible():
                self.focusSwitch.emit(self.parent.recentList.count()-1)
            else:
                self._move(1)

        elif key == QtCore.Qt.Key_Down:
            if self.parent.recentList.isVisible():
                self.focusSwitch.emit(0)
            else:
                self._move(-1)

        elif key == QtCore.Qt.Key_Backspace and event.modifiers() == QtCore.Qt.ShiftModifier:
            self.clear()
            self.textEdited.emit("")

        self.backspace = key == QtCore.Qt.Key_Backspace
        QtWidgets.QLineEdit.keyPressEvent(self, event)

    def _move(self, direction):
        self.clearRecent.emit()
        self.index += direction
        length = len(self.parent.history)
        if self.index == length or self.index == -length:
            self.index = 0
        self.setText(self.parent.history[self.index])

    def _setDefaultPalette(self):
        self.setPalette(self.parent.palette())
        self.setAttribute(QtCore.Qt.WA_SetPalette, False)

    def updateText(self, row):
        item = self.parent.recentList.item(row)
        if item:
            self.setText(item.text())


class RecentList(QtWidgets.QListWidget):
    removeSuggestion = QtCore.pyqtSignal(str)
    execute = QtCore.pyqtSignal(object)
    focusCmd = QtCore.pyqtSignal(bool)
    focusOut = QtCore.pyqtSignal()
    deselect = QtCore.pyqtSignal()
    resize = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.itemClicked.connect(self._itemClicked)
        self.resizeToContent()

    def keyPressEvent(self, event):
        self._setDefaultPalette()
        key = event.key()

        if key == QtCore.Qt.Key_Return or key == QtCore.Qt.Key_Enter:
            self.execute.emit(self.currentItem())

        elif key == QtCore.Qt.Key_Delete:
            self.removeSuggestion.emit(self.currentItem().text())
            self.takeItem(self.currentRow())
            self.resizeToContent()
            self.resize.emit()

        elif key == QtCore.Qt.Key_Down:
            if self.currentRow() == self.count()-1:
                self._switchFocus()

        elif key == QtCore.Qt.Key_Up:
            if self.currentRow() == 0:
                self._switchFocus()

        else:
            self._switchFocus()
            self.deselect.emit()
            self._postKey(event)
        QtWidgets.QListWidget.keyPressEvent(self, event)

    def _clearSelection(self):
        for i in range(self.count()):
            self.item(i).setSelected(False)

    def _itemClicked(self, item):
        self.execute.emit(item)

    def _postKey(self, event):
        try:
            sequence = QtGui.QKeySequence(event.key()).toString().lower()
            sequence = sequence.replace("space", " ")
            sequence.encode('utf-8')
        except UnicodeEncodeError:
            # Filter unwanted characters from modifiers
            sequence = None
        keyEvent = QtGui.QKeyEvent(QtCore.QEvent.KeyPress, event.key(), event.modifiers(), sequence)
        QtCore.QCoreApplication.postEvent(self.parent.cmdLine, keyEvent)

    def _setDefaultPalette(self):
        self.setPalette(self.parent.palette())
        self.setAttribute(QtCore.Qt.WA_SetPalette, False)

    def _switchFocus(self):
        self._clearSelection()
        self.focusCmd.emit(True)

    def focusOutEvent(self, event):
        self.focusOut.emit()

    def resizeToContent(self):
        for row in range(self.count()):
            item = self.item(row)
            item.setSizeHint(QtCore.QSize(item.sizeHint().width(), 18))

        height = self.sizeHintForRow(0) * self.count()
        if height:
            self.setMinimumHeight(height + 2)
            self.setVisible(True)
        else:
            self.setVisible(False)
