#!/usr/bin/python3
import os
import sys
from PyQt5 import QtGui, QtWidgets, QtCore, QtMultimedia, QtDBus

try:
    from obsuite.backend.common import Slave, trayEventParser
except ImportError:
    from backend.common import Slave, trayEventParser

LOCAL_DIR = os.path.dirname(os.path.realpath(__file__)) + '/'
TIMEOUT = 6000 ##


class Status(QtCore.QObject):
    popup = QtCore.pyqtSignal(str)
    fetch = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()
        self.slave = parent.slave
        self.ready = False
        self.mute = False
        self.lastMute = False
        self.lastVolume = -1
        self.level = 0

    def _parse(self, output):
        for line in output.splitlines():
            if line.find("Front Left: Playback") > -1 or line.find("Front Right: Playback") > -1:
                level = line[line.find("[")+1 : line.find("]")]
                level = level.replace("%","")
                self.level = int(level)
                if line[line.find("] [")+3:-1] == "off":
                    self.mute = True
                else:
                    self.mute = False

    def update(self, output):
        self._parse(output)
        if self.level == 0 or self.mute:
            self.icon = "mute"
        elif self.level > 66:
            self.icon = "high"
        elif self.level > 33:
            self.icon = "medium"
        elif self.level > 0:
            self.icon = "low"

        if not self.lastVolume == self.level:
            if self.lastVolume > -1:
                self.popup.emit(f"{self.level}%")
        elif not self.lastMute == self.mute:
            self.popup.emit("Mute enabled" if self.mute else "Mute disabled")

        self.lastVolume = self.level
        self.lastMute = self.mute
        self.ready = True
        self.fetch.emit()


class Bluez(QtCore.QObject):
    popupQueued = QtCore.pyqtSignal(str)
    togglingDone = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.preferences = parent.preferences
        self.log = parent.parent.logger.new("Bluetooth")
        self.lastStatus = False
        self.setInterfaces()
        self.statusTimer = QtCore.QTimer(interval=1000)
        self.statusTimer.timeout.connect(self._statusQuery)
        self.statusTimer.start()

    def setInterfaces(self):
        self.reply = []
        self.devices = {}
        self.properties = {}
        for device in self.preferences.get("parameters", "bluetooth"):
            self._addInterface(device)

    def _addInterface(self, path):
        service = "org.bluez"
        bus = QtDBus.QDBusConnection.systemBus()
        self.properties[path] = QtDBus.QDBusInterface(service, path, "org.freedesktop.DBus.Properties", bus)
        self.devices[path] = QtDBus.QDBusInterface(service, path, "org.bluez.Device1", bus)
        self.devices[path].setTimeout(TIMEOUT)

    def _connectAttempt(self):
        for path in self.devices:
            name = self.preferences.get("parameters", "bluetooth", path)
            self.popupQueued.emit(f"Connecting to...\n{name}")
            reply = self.devices[path].call("Connect")
            error = reply.arguments()[0]
            if error:
                self.log.info(f"Connection failed: {error}")
            else:
                self.log.info("Connection successful")
                break

    def _statusQuery(self):
        for path in self.properties:
            self.properties[path].callWithCallback("Get", ["org.bluez.MediaControl1", "Connected"], self.slotStatus)

    def _statusCompare(self):
        self.status = sum(self.reply)
        if not self.lastStatus == self.status:
            if self.status > self.lastStatus:
                self.popupQueued.emit("Bluetooth connected")
            else:
                self.popupQueued.emit("Bluetooth disconnected")
            self.lastStatus = self.status
        self.reply = []

    def connect(self, path):
        name = self.preferences.get("parameters", "bluetooth", path)
        self.popupQueued.emit(f"Connecting to...\n{name}")
        self.devices[path].asyncCall("Connect")

    def disconnect(self, path):
        self.devices[path].asyncCall("Disconnect")

    @QtCore.pyqtSlot(QtDBus.QDBusMessage)
    def slotStatus(self, reply):
        self.reply.append(reply.arguments()[0])
        if len(self.reply) == len(self.devices):
            self._statusCompare()

    @QtCore.pyqtSlot()
    def toggleConnection(self):
        for path in self.properties:
            # Disconnect active bluetooth device
            reply = self.properties[path].call("Get", "org.bluez.MediaControl1", "Connected")
            connected = reply.arguments()[0]
            if connected:
                name = self.preferences.get("parameters", "bluetooth", path)
                self.devices[path].asyncCall("Disconnect")
                self.log.info(f"Called 'Disconnect' on '{name}'")
                break
        else:
            # Connect to the first available bluetooth device
            self._connectAttempt()
        self.togglingDone.emit()


class Main(QtWidgets.QSystemTrayIcon):
    popupQueued = QtCore.pyqtSignal(object, str)
    popup = QtCore.pyqtSignal(object, str)
    error = QtCore.pyqtSignal(object, int)
    execute = QtCore.pyqtSignal(str)

    def __init__(self, parent):
        super().__init__()
        self.name = "volume"
        self.parent = parent
        self.preferences = parent.preferences
        self.slave = Slave(self)
        self.status = Status(self)
        self.fetchTimer = QtCore.QTimer(singleShot=True)

        self.bluezToggling = False
        self.bluez = Bluez(self)
        self.bluezThread = QtCore.QThread()
        self.bluez.moveToThread(self.bluezThread)
        self.bluez.togglingDone.connect(self._bluezTogglingDone)
        self.bluezThread.start()

        self.slave.ready.connect(self._update)
        self.activated.connect(self._clickEvent)
        self.bluez.popupQueued.connect(self._popupQueued)
        self.installEventFilter(self)
        self.status.popup.connect(self._popup)
        self.popup.connect(parent.display)
        self.popupQueued.connect(parent.displayQueued)

        self.setIcon(QtGui.QIcon(f"{LOCAL_DIR}../icons/{self.name}/default.svg"))
        if self.preferences.get("parameters", "volume", "tray icon"):
            self.show()

    def _bluezTogglingDone(self):
        self.bluezToggling = False

    def _clickEvent(self, event):
        self._eventParser(event)

    def _error(self, code):
        self.error.emit(self, code)

    def _eventParser(self, event):
        trigger = trayEventParser(event)
        if trigger:
            enabled = self.preferences.get("actions", self.name, trigger)[0]
            if enabled:
                action = self.preferences.get("actions", self.name, trigger)[1].lower()
                command = self.preferences.get("exec", self.name, trigger)
                self.action(action, command)

    def _fetch(self):
        self.slave.fetch("amixer scontents")

    def _popup(self, message):
        self.popup.emit(self, message)

    def _popupQueued(self, message):
        self.popupQueued.emit(self, message)

    def _update(self, output):
        self.status.update(output)
        self.setToolTip(f"Volume: {self.status.level} %")
        self.setIcon(self.icons[self.status.icon])

    def action(self, action, command=""):
        if action == "exec":
            self.execute.emit(command)

        elif action == "toggle sound level":
            if self.status.level >= self.preferences.get("parameters", "volume", "threshold"):
                self.slave.start(self.preferences.get("parameters", "volume", "low command"))
            else:
                self.slave.start(self.preferences.get("parameters", "volume", "high command"))
            self.slave.waitForFinished()

        elif action == "toggle bluetooth":
            print(self.bluezToggling)
            if not self.bluezToggling:
                self.bluezToggling = True
                QtCore.QMetaObject.invokeMethod(self.bluez, "toggleConnection", QtCore.Qt.QueuedConnection)

        else:
            self.log.error(f"Invalid action '{action}'")

    def eventFilter(self, object, event):
        self._eventParser(event)
        return QtWidgets.QSystemTrayIcon.event(self, event)

    def initMenu(self):
        self.menu = QtWidgets.QMenu()
        for device in self.preferences.get("parameters", "bluetooth"):
            name = self.preferences.get("parameters", "bluetooth", device)
            self.menu.addAction(self.icons["connect"], f"Connect {name}", lambda device=device: self.bluez.connect(device))
            self.menu.addAction(self.icons["disconnect"], f"Disconnect {name}", lambda device=device: self.bluez.disconnect(device))
            self.menu.addSeparator()
        self.setContextMenu(self.menu)
