#!/usr/bin/python3
import copy
import os
import shlex
import subprocess
import sys
from PyQt5 import QtGui, QtWidgets, QtCore, uic

try:
    import obsuite.backend.logger as logger
    import obsuite.backend.qdbus as qdbus
    import obsuite.backend.bus as bus
    import obsuite.frontend.preferences as preferences
    import obsuite.frontend.popup as popup
    import obsuite.frontend.battery as battery
    import obsuite.frontend.volume as volume
    import obsuite.frontend.brightness as brightness
    import obsuite.frontend.network as network
    import obsuite.frontend.temperature as temperature
    import obsuite.frontend.launcher as launcher
    import obsuite.frontend.splash as splash
except ImportError:
    import backend.logger as logger
    import backend.qdbus as qdbus
    import backend.bus as bus
    import frontend.preferences as preferences
    import frontend.popup as popup
    import frontend.battery as battery
    import frontend.volume as volume
    import frontend.brightness as brightness
    import frontend.network as network
    import frontend.temperature as temperature
    import frontend.launcher as launcher
    import frontend.splash as splash

LOCAL_DIR = os.path.dirname(os.path.realpath(__file__)) + '/'
CONFIG_DIR = os.path.expanduser("~/.config/obsuite/")


class Main:
    def __init__(self, parent, waitForSystray=False):
        super().__init__()
        self.parent = parent
        self.logger = logger.Logger(name="obsuite", path=CONFIG_DIR)
        self.logExec = self.logger.new("Execute")
        self.log = self.logger.new("Main")

        if waitForSystray:
            import time
            self.log.info("Waiting for system tray...")
            while not QtWidgets.QSystemTrayIcon.isSystemTrayAvailable():
                time.sleep(1)
            time.sleep(1)

        self.icons = {}
        for widget in ("battery", "brightness", "volume", "network", "temperature", "bus"):
            path = f"{LOCAL_DIR}icons/{widget}"
            for icon in os.listdir(path):
                self.icons[f"{widget}/{icon}"] = QtGui.QIcon(f"{path}/{icon}")
        self.process = QtCore.QProcess()
        self.preferences = preferences.PreferencesDatabase(self)
        self.preferencesForm = preferences.PreferencesForm(self)
        self.preferencesForm.accepted.connect(self.preferencesSave)
        self.popup = popup.PopupWidget(self)
        self.popup.done.connect(self._setIdle)
        self._load()

    def _load(self):
        self.actives = []
        if self.preferences.get("parameters", "battery", "enable"):
            self.battery = battery.Main(self)
            self.battery.execute.connect(self.execute)
            self.actives.append(self.battery)

        if self.preferences.get("parameters", "volume", "enable"):
            self.volume = volume.Main(self)
            self.volume.execute.connect(self.execute)
            self.actives.append(self.volume)

        if self.preferences.get("parameters", "brightness", "enable"):
            self.brightness = brightness.Main(self)
            self.brightness.execute.connect(self.execute)
            self.actives.append(self.brightness)

        if self.preferences.get("parameters", "temperature", "enable"):
            self.temperature = temperature.Main(self)
            self.actives.append(self.temperature)

        if self.preferences.get("parameters", "network", "enable"):
            self.network = network.Main(self)
            self.actives.append(self.network)

        for widget in self.actives:
            widget.icons = {}
            for icon in os.listdir(f"{LOCAL_DIR}icons/{widget.name}"):
                widget.icons[f"{os.path.splitext(icon)[0]}"] = QtGui.QIcon(f"{LOCAL_DIR}icons/{widget.name}/{icon}")
            widget.log = self.logger.new(widget.name.capitalize())
            widget.error.connect(self.error)
            widget.status.fetch.connect(widget.fetchTimer.start)
            widget.slave.ready.connect(widget.status.update)
            widget.slave.errorOccurred.connect(widget._error)
            widget.fetchTimer.timeout.connect(widget._fetch)
            widget.fetchTimer.setInterval(500)
            widget.fetchTimer.start()

        self.bus = bus.Main(self)
        self.bus.execute.connect(self.execute)

        if self.preferences.get("parameters", "volume", "enable"):
            self.volume.initMenu()

    def _reload(self):
        for widget in self.actives:
            widget.deleteLater()
        self._load()

    def _setIdle(self):
        for widget in self.actives:
            widget.fetchTimer.setInterval(500)

    def display(self, widget, message):
        if widget.status.ready:
            self.popup.display(widget, message)
        message = message.replace("\n", ", ")
        widget.log.info(f"{message} (popup)")
        widget.fetchTimer.setInterval(100)

    def displayQueued(self, widget, message):
        if widget.status.ready:
            self.popup.queue(widget, message)
        message = message.replace("\n", ", ")
        widget.log.info(f"{message} (queued popup)")

    def error(self, widget, error):
        reason = \
        {
            QtCore.QProcess.FailedToStart : "Failed to start",
            QtCore.QProcess.Crashed : "Crashed",
            QtCore.QProcess.Timedout : "Timedout",
            QtCore.QProcess.WriteError : "Write error",
            QtCore.QProcess.ReadError : "Read error",
            QtCore.QProcess.UnknownError : "Unknown error"
        }
        if error in reason:
            error = f"Process error:\n{reason[error]}"
            widget.fetchTimer.stop()
            if isinstance(widget, QtWidgets.QSystemTrayIcon):
                widget.hide()
        widget.status.icon = "critical"
        self.popup.queue(widget, f"{error}")
        error = error.replace("\n", " ")
        widget.log.critical(f"{error}")

    def execute(self, cmd):
        if cmd:
            cmd = cmd.replace("~", os.path.expanduser("~"))
            try:
                subprocess.Popen(shlex.split(cmd), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                self.logExec.info(cmd)
            except FileNotFoundError:
                self.logExec.error(f"Invalid command '{cmd}'")

    def parseCommands(self, cmd):
        if "quit" in cmd:
            self.parent.exit()

        elif "preferences" in cmd:
            if self.preferencesForm.isVisible():
                self.preferencesForm.hide()
            else:
                self.preferencesForm.show()
                self.preferencesForm.raise_()

        elif "run" in cmd:
            launcher.Main(self)

        elif "splash" in cmd:
            s = splash.Main(self)
            s.execute.connect(self.execute)

        elif "reload" in cmd:
            self._reload()

        elif "action" in cmd:
            if len(cmd["action"]) > 1:
                args = iter(cmd["action"])
                widget, action = next(args), next(args)
                command = next(args, "")
                if widget == "battery" and hasattr(self, "battery"):
                    self.battery.action(action, command)
                elif widget == "volume" and hasattr(self, "volume"):
                    self.volume.action(action, command)
                elif widget == "brightness" and hasattr(self, "brightness"):
                    self.brightness.action(action, command)
            else:
                self.log.error('Correct format: obsuite --action widget "action"')


    def preferencesActivesChanged(self):
        widgets = ("battery", "brightness", "network", "temperature", "volume")
        for w in widgets:
            old = self.preferences.db["parameters"][w]["enable"]
            new = self.preferencesForm.db["parameters"][w]["enable"]
            if not new == old:
                return True
        return False

    def preferencesSave(self):
        self.preferencesForm.settingsSave()
        refresh = self.preferencesActivesChanged()
        self.preferences.db = copy.deepcopy(self.preferencesForm.db)
        self.preferences.save()
        if refresh:
            self._reload()
        else:
            self.updateTrayIcons()
            self.updateBluezDevices()
            self.popup.updateView()

    def updateBluezDevices(self):
        if hasattr(self, "volume"):
            self.volume.bluez.setInterfaces()
            self.volume.initMenu()

    def updateTrayIcons(self):
        for widget in self.actives:
            if widget.name in ("battery", "brightness", "volume"):
                visible = self.preferences.get("parameters", widget.name, "tray icon")
                widget.setVisible(visible)


def main(cmd):
    app = QtWidgets.QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(False)
    waitForSystray = cmd.lower().count("--wait-for-systray")
    daemon = Main(app, waitForSystray)
    daemon.parseCommands(cmd)
    bus = qdbus.QDBusObject(parent=daemon)
    sys.exit(app.exec_())
